/**************************************************************************
*  Copyright (c) 2013 by Michael Fischer. All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
*  This IRQ handler based on source from Rowley and Giovanni Di Sirio.
*
*  Copyright (c) 2012 Rowley Associates Limited.
*  Copyright (c) 2006-2012, Giovanni Di Sirio.
*
***************************************************************************
*  History:
*
*  09.02.2013  mifi  First Version for the BeagleBone board.
*                    The IRQ handler is based on the original version
*                    from CrossWorks for ARM v2.3.0, Sitara BSP v1.0.
*                    The handler was expanded by the ChibiOS functionality.
**************************************************************************/

#include "chconf.h"

#define FALSE 0
#define TRUE 1

/*
 * Common exit point for all IRQ routines, it performs the rescheduling if
 * required.
 * System stack frame structure after a context switch in the
 * interrupt handler:
 *
 * High +------------+
 *      |   LR_USR   | -+
 *      |     R12    |  |
 *      |     R3     |  |
 *      |     R2     |  | External context: IRQ handler frame
 *      |     R1     |  |
 *      |     R0     |  |
 *      |     PC     |  |   (user code return address)
 *      |   PSR_USR  | -+   (user code status)
 *      |    ....    | <- chSchDoReschedule() stack frame, optimize it for space
 *      |     LR     | -+   (system code return address)
 *      |     R11    |  |
 *      |     R10    |  |
 *      |     R9     |  |
 *      |     R8     |  | Internal context: chSysSwitch() frame
 *      |    (R7)    |  |   (optional, see CH_CURRP_REGISTER_CACHE)
 *      |     R6     |  |
 *      |     R5     |  |
 * SP-> |     R4     | -+
 * Low  +------------+
 */


/*
 * This is the IRQ handler, based on the CrossWorks for ARM version.
 * Expanded by the ChibiOS/RT functionality. This handler will
 * override the weak version from CrossWorks. 
 */
.text
.balign 16
.code 32

.set    MODE_IRQ, 0x12
.set    MODE_SVC, 0x13
.set    MODE_SYS, 0x1F
.equ    I_BIT, 0x80

#include "AM335x.h"
#define MASK_ACTIVE_IRQ     0x7f
#define NEWIRQAGR           1

.globl irq_handler
irq_handler:

        // Store the APCS registers
        sub lr, lr, #4
        push {r0-r3, r12, lr}
  
#ifdef PRESERVE_VFP_REGISTERS
        vpush {d0-d7}
        vpush {d16-d31}
        vmrs r0, fpscr
        push {r0}  
#endif

        // Access registers via the base
        ldr r0, =INTC_BASE

        // Save the current threshold priority  
        ldr r1, [r0, #INTC_THRESHOLD_OFFSET]             
        push {r1}
 
        // Update the threshold to that of the active irq
        ldr r3, [r0, #INTC_IRQ_PRIORITY_OFFSET]
        str r3, [r0, #INTC_THRESHOLD_OFFSET]

        // Get the active irq in r2
        ldr r2, [r0, #INTC_SIR_IRQ_OFFSET]
        and r2, r2, #MASK_ACTIVE_IRQ

        // Acknowledge irq if priority is not zero
        mov r1, #NEWIRQAGR
        cmp r3, #0
        strne r1, [r0, #INTC_CONTROL_OFFSET]
        dsb

        // r2 - irq number, r3 - irq priority

        // Switch to system mode with interrupts enabled if priority is not zero
        mrs  r12, spsr
        push {r12}

        mrs r14, cpsr
        orr r14, r14, #MODE_SYS
        
#if 0   // mifi: Do not enable the IRQ
        cmp r3, #0
        bicne r14, r14, #I_BIT
#endif

        msr cpsr, r14       
          
        // Now running in system mode - need to ensure that sp is 8 byte aligned to call the ISR
        mov r1, sp
        and r1, r1, #4
        sub sp, sp, r1
        // Save the system mode lr and the sp adjustment
        push {r1, lr}

        // Call the ISR
        ldr r0, =fnRAMVectors
        mov lr, pc
        ldr pc, [r0, r2, lsl #2]

        // Restore the system mode lr and remove the sp adjustment
        pop {r1, lr}
        add sp, sp, r1

        // Switch back to irq mode
        cpsid i, #MODE_IRQ
        pop {r3}
        msr spsr, r3 

        // Access registers via the base
        ldr r0, =INTC_BASE      

        // Acknowledge irq if priority is not zero
        ldr r1, [r0, #INTC_THRESHOLD_OFFSET]
        cmp r1, #0
        moveq r2, #NEWIRQAGR
        streq r2, [r0, #INTC_CONTROL_OFFSET]

        // Restore threshold
        pop {r1}
        str r1, [r0, #INTC_THRESHOLD_OFFSET]
    
        //  
        // mifi: Check if scheduling is required  
        //
        bl  chSchIsPreemptionRequired
        cmp r0, #0
        bne irq_handler_schedule  // Scheduling is required
                           
        //
        // This is the normal IRQ end
        //                                            
#ifdef PRESERVE_VFP_REGISTERS
        pop {r0}
        vmsr fpscr, r0
        vpop {d16-d31}
        vpop {d0-d7}  
#endif

        // Return from interrupt
        pop {r0-r3, r12, lr}
        movs pc, lr
        ///////////////////////////////////////////////
  
        //
        // Scheduling is required
        //  
irq_handler_schedule:

        // Saves the IRQ mode registers in the system stack.
        ldmfd   sp!, {r0-r3, r12, lr}           // IRQ stack now empty.
        msr     CPSR_c, #MODE_SYS | I_BIT
        stmfd   sp!, {r0-r3, r12, lr}           // Registers on System Stack.
        msr     CPSR_c, #MODE_IRQ | I_BIT
        mrs     r0, SPSR
        mov     r1, lr
        msr     CPSR_c, #MODE_SYS | I_BIT
        stmfd   sp!, {r0, r1}                   // Push R0=SPSR, R1=LR_IRQ.

        // Context switch.
#if CH_DBG_SYSTEM_STATE_CHECK
        bl      dbg_check_lock
#endif

        bl      chSchDoReschedule
        
#if CH_DBG_SYSTEM_STATE_CHECK
        bl      dbg_check_unlock
#endif

        // Re-establish the IRQ conditions again.
        ldmfd   sp!, {r0, r1}                   // Pop R0=SPSR, R1=LR_IRQ.
        msr     CPSR_c, #MODE_IRQ | I_BIT
        msr     SPSR_fsxc, r0
        mov     lr, r1
        msr     CPSR_c, #MODE_SYS | I_BIT
        ldmfd   sp!, {r0-r3, r12, lr}
        msr     CPSR_c, #MODE_IRQ | I_BIT
        movs    pc, lr


loop_end:
  b loop_end

/*** EOF ***/

